##############################################################################
#
# DavPack -- A set of patches for all things related to WebDAV
# Copyright (C) 2004 Enfold Systems
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Some portions of this module are Copyright Zope Corporatioin and
# Contributors.
# The original copyright statement is reproduced below.
#
##############################################################################
#
# Copyright (c) 2001 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
$Id: __init__.py 729 2005-05-27 19:14:11Z sidnei $
"""

import sys
import OFS
from zLOG import LOG, INFO

def log(msg):
    LOG('DavPack', INFO, msg)

from Acquisition import aq_base
from webdav.WriteLockInterface import WriteLockInterface
from webdav import Lockable
from webdav.common import Locked, Conflict, PreconditionFailed
from webdav.common import absattr, isDavCollection
from webdav.davcmds import PropPatch, Lock
import webdav.Resource, webdav.Collection
from OFS import PropertySheets
from ComputedAttribute import ComputedAttribute
from zExceptions import Unauthorized, Forbidden
from zExceptions import BadRequest, MethodNotAllowed

_marker = []

def site_encoding(ctx, default='utf-8'):
    from Products.CMFCore.utils import getToolByName
    pprops = getToolByName(ctx, 'portal_properties', None)
    if pprops is not None:
        sprops = getToolByName(pprops, 'site_properties', None)
        if sprops is not None:
            encoding = sprops.getProperty('default_charset')
            if encoding is not None:
                return encoding
    return default

# This is the same value used in zhttp_collector. It should turn into
# a ZConfig setting on zope.conf when it gets merged into Zope.
LARGE_FILE = 1 << 19

def PUT(self, REQUEST, RESPONSE):
    """Create a new non-collection resource."""
    self.dav__init(REQUEST, RESPONSE)

    name = self.__name__
    parent = self.__parent__

    ifhdr = REQUEST.get_header('If', '')
    if WriteLockInterface.isImplementedBy(parent) and parent.wl_isLocked():
        if ifhdr:
            parent.dav__simpleifhandler(REQUEST, RESPONSE, col=1)
        else:
            # There was no If header at all, and our parent is locked,
            # so we fail here
            raise Locked
    elif ifhdr:
        # There was an If header, but the parent is not locked
        raise PreconditionFailed

    # SDS: Only use BODY if the file size is smaller than LARGE_FILE,
    # otherwise read LARGE_FILE bytes from the file which should be
    # enough to trigger content_type detection, and possibly enough
    # for CMF's content_type_registry too.
    #
    # Note that body here is really just used for detecting the
    # content type and figuring out the correct factory. The correct
    # file content will be uploaded on ob.PUT(REQUEST, RESPONSE) after
    # the object has been created.
    #
    # A problem I could see is content_type_registry predicates that
    # do depend on the whole file being passed here as an
    # argument. There's none by default that does this though. If they
    # really do want to look at the file, they should use
    # REQUEST['BODYFILE'] directly.

    if int(REQUEST.get('CONTENT_LENGTH') or 0) > LARGE_FILE:
        file = REQUEST['BODYFILE']
        body = file.read(LARGE_FILE)
        file.seek(0)
    else:
        body = REQUEST.get('BODY', '')

    typ = REQUEST.get_header('content-type', None)
    if typ is None:
        typ, enc = OFS.content_types.guess_content_type(name, body)

    factory = getattr(parent, 'PUT_factory', self._default_PUT_factory)
    ob = factory(name, typ, body)
    if ob is None:
        ob = self._default_PUT_factory(name, typ, body)

    # We call _verifyObjectPaste with verify_src=0, to see if the
    # user can create this type of object (and we don't need to
    # check the clipboard.
    try:
        parent._verifyObjectPaste(ob.__of__(parent), 0)
    except Unauthorized:
        raise
    except:
        raise Forbidden, sys.exc_info()[1]

    # Delegate actual PUT handling to the new object,
    # SDS: But just *after* it has been stored.
    self.__parent__._setObject(name, ob)
    # Give the object a _p_jar, required for some operations like
    # renaming, and in theory harmless at this point in the call
    # stack.
    get_transaction().commit(1)
    ob = self.__parent__._getOb(name)
    ob.PUT(REQUEST, RESPONSE)

    RESPONSE.setStatus(201)
    RESPONSE.setBody('')
    return RESPONSE


def COPY(self, REQUEST, RESPONSE):
    """Create a duplicate of the source resource whose state
    and behavior match that of the source resource as closely
    as possible. Though we may later try to make a copy appear
    seamless across namespaces (e.g. from Zope to Apache), COPY
    is currently only supported within the Zope namespace.
    """
    self.dav__init(REQUEST, RESPONSE)
    if not hasattr(aq_base(self), 'cb_isCopyable') or \
       not self.cb_isCopyable():
        raise MethodNotAllowed, 'This object may not be copied.'

    depth=REQUEST.get_header('Depth', 'infinity')
    if not depth in ('0', 'infinity'):
        raise BadRequest, 'Invalid Depth header.'

    dest=REQUEST.get_header('Destination', '')
    while dest and dest[-1]=='/':
        dest=dest[:-1]
    if not dest:
        raise BadRequest, 'Invalid Destination header.'

    try: path = REQUEST.physicalPathFromURL(dest)
    except ValueError:
        raise BadRequest, 'Invalid Destination header'

    name = path.pop()
    parent_path = '/'.join(path)

    oflag=REQUEST.get_header('Overwrite', 'F').upper()
    if not oflag in ('T', 'F'):
        raise BadRequest, 'Invalid Overwrite header.'

    try: parent=self.restrictedTraverse(path)
    except ValueError:
        raise Conflict, 'Attempt to copy to an unknown namespace.'
    except NotFound:
        raise Conflict, 'Object ancestors must already exist.'
    except:
        t, v, tb=sys.exc_info()
        raise t, v
    if hasattr(parent, '__null_resource__'):
        raise Conflict, 'Object ancestors must already exist.'
    # ---- Start Plone Hack ---------------------------------------
    # Plone-specific hack that can't just go into Zope. Can be removed
    # when ReplaceableWrapper has been removed from Plone.
    existing = True
    obj = getattr(aq_base(parent), name, _marker)
    if obj is _marker or isinstance(obj, ComputedAttribute):
        existing = False
    # ---- End Plone Hack ---------------------------------------
    if existing and oflag=='F':
        raise PreconditionFailed, 'Destination resource exists.'
    try:
        parent._checkId(name, allow_dup=1)
    except:
        raise Forbidden, sys.exc_info()[1]
    try:
        parent._verifyObjectPaste(self)
    except Unauthorized:
        raise
    except:
        raise Forbidden, sys.exc_info()[1]

    # Now check locks.  The If header on a copy only cares about the
    # lock on the destination, so we need to check out the destinations
    # lock status.
    ifhdr = REQUEST.get_header('If', '')
    if existing:
        # The destination itself exists, so we need to check its locks
        destob = aq_base(parent)._getOb(name)
        if WriteLockInterface.isImplementedBy(destob) and \
           destob.wl_isLocked():
            if ifhdr:
                itrue = destob.dav__simpleifhandler(
                    REQUEST, RESPONSE, 'COPY', refresh=1)
                if not itrue:
                    raise PreconditonFailed
            else:
                raise Locked, 'Destination is locked.'
    elif WriteLockInterface.isImplementedBy(parent) and \
         parent.wl_isLocked():
        if ifhdr:
            parent.dav__simpleifhandler(REQUEST, RESPONSE, 'COPY',
                                        refresh=1)
        else:
            raise Locked, 'Destination is locked.'

    ob = self._getCopy(parent)
    ob._setId(name)

    if depth=='0' and isDavCollection(ob):
        for id in ob.objectIds():
            ob._delObject(id)
    if existing:
        object=getattr(parent, name)
        self.dav__validate(object, 'DELETE', REQUEST)
        parent._delObject(name)
    parent._setObject(name, ob)
    ob = parent._getOb(name)
    ob.manage_afterClone(ob)
    # We remove any locks from the copied object because webdav clients
    # don't track the lock status and the lock token for copied resources
    ob.wl_clearLocks()
    RESPONSE.setStatus(existing and 204 or 201)
    if not existing:
        RESPONSE.setHeader('Location', dest)
    RESPONSE.setBody('')
    return RESPONSE


def MOVE(self, REQUEST, RESPONSE):
    """Move a resource to a new location. Though we may later try to
    make a move appear seamless across namespaces (e.g. from Zope
    to Apache), MOVE is currently only supported within the Zope
    namespace."""
    self.dav__init(REQUEST, RESPONSE)
    self.dav__validate(self, 'DELETE', REQUEST)
    if not hasattr(aq_base(self), 'cb_isMoveable') or \
       not self.cb_isMoveable():
        raise MethodNotAllowed, 'This object may not be moved.'

    dest=REQUEST.get_header('Destination', '')

    try: path = REQUEST.physicalPathFromURL(dest)
    except ValueError:
        raise BadRequest, 'No destination given'

    flag=REQUEST.get_header('Overwrite', 'F')
    flag=flag.upper()

    name = path.pop()
    parent_path = '/'.join(path)

    try: parent = self.restrictedTraverse(path)
    except ValueError:
        raise Conflict, 'Attempt to move to an unknown namespace.'
    except 'Not Found':
        raise Conflict, 'The resource %s must exist.' % parent_path
    except:
        t, v, tb=sys.exc_info()
        raise t, v
    if hasattr(parent, '__null_resource__'):
        raise Conflict, 'The resource %s must exist.' % parent_path
    # ---- Start Plone Hack ---------------------------------------
    # Plone-specific hack that can't just go into Zope. Can be removed
    # when ReplaceableWrapper has been removed from Plone.
    existing = True
    obj = getattr(aq_base(parent), name, _marker)
    if obj is _marker or isinstance(obj, ComputedAttribute):
        existing = False
    # ---- End Plone Hack ---------------------------------------
    if existing and flag=='F':
        raise PreconditionFailed, 'Resource %s exists.' % dest
    try:
        parent._checkId(name, allow_dup=1)
    except:
        raise Forbidden, sys.exc_info()[1]
    try:
        parent._verifyObjectPaste(self)
    except Unauthorized:
        raise
    except:
        raise Forbidden, sys.exc_info()[1]

    # Now check locks.  Since we're affecting the resource that we're
    # moving as well as the destination, we have to check both.
    ifhdr = REQUEST.get_header('If', '')
    if existing:
        # The destination itself exists, so we need to check its locks
        destob = aq_base(parent)._getOb(name)
        if WriteLockInterface.isImplementedBy(destob) and \
           destob.wl_isLocked():
            if ifhdr:
                itrue = destob.dav__simpleifhandler(
                    REQUEST, RESPONSE, 'MOVE', url=dest, refresh=1)
                if not itrue:
                    raise PreconditionFailed
            else:
                raise Locked, 'Destination is locked.'
    elif WriteLockInterface.isImplementedBy(parent) and \
         parent.wl_isLocked():
        # There's no existing object in the destination folder, so
        # we need to check the folders locks since we're changing its
        # member list
        if ifhdr:
            itrue = parent.dav__simpleifhandler(REQUEST, RESPONSE, 'MOVE',
                                                col=1, url=dest, refresh=1)
            if not itrue:
                raise PreconditionFailed, 'Condition failed.'
        else:
            raise Locked, 'Destination is locked.'
    if Lockable.wl_isLocked(self):
        # Lastly, we check ourselves
        if ifhdr:
            itrue = self.dav__simpleifhandler(REQUEST, RESPONSE, 'MOVE',
                                              refresh=1)
            if not itrue:
                raise PreconditionFailed, 'Condition failed.'
        else:
            raise PreconditionFailed, 'Source is locked and no '\
                  'condition was passed in.'

    ob=aq_base(self._getCopy(parent))
    self.aq_parent._delObject(absattr(self.id))
    ob._setId(name)
    if existing:
        object=getattr(parent, name)
        self.dav__validate(object, 'DELETE', REQUEST)
        parent._delObject(name)
    parent._setObject(name, ob)
    RESPONSE.setStatus(existing and 204 or 201)
    if not existing:
        RESPONSE.setHeader('Location', dest)
    RESPONSE.setBody('')
    return RESPONSE


def xml_escape(value):
    from webdav.xmltools import escape
    if not isinstance(value, basestring):
        value = str(value)
    value = escape(value)
    return value

old_allprop = PropertySheets.PropertySheet.dav__allprop
def dav__allprop(self, propstat=old_allprop.func_defaults[-1]):
    enc = site_encoding(self, default='latin-1')
    result = old_allprop(self, propstat=propstat)
    if isinstance(result, str):
        # decode into the portal encoding.
        result = result.decode(enc)
    # xml output will be utf-8
    return result.encode('utf-8')

old_propstat = PropertySheets.PropertySheet.dav__propstat
def dav__propstat(self, name, result,
                  propstat=old_propstat.func_defaults[-2],
                  propdesc=old_propstat.func_defaults[-1]):
    enc = site_encoding(self, default='latin-1')
    old_propstat(self, name, result,
                 propstat=propstat, propdesc=propdesc)
    for key, items in result.items():
        for idx, value in enumerate(items):
            if isinstance(value, str):
                # decode into the portal encoding and encode into utf-8
                # xml output will be utf-8
                items[idx] = value.decode(enc).encode('utf-8')
    return

old_PropPatch_apply = PropPatch.apply
def PropPatch_apply(self, obj):
    enc = site_encoding(obj, default='latin-1')
    # strval() has been patched to return unicode strings, so we
    # re-encode into site-encoding for the object.
    for idx, value in enumerate(self.values):
        if len(value) > 2:
            name, ns, val, md = value
            if isinstance(val, unicode):
                val = val.encode(enc)
            if isinstance(name, unicode):
                name = name.encode(enc)
            if isinstance(ns, unicode):
                ns = ns.encode(enc)
            if md.has_key('__xml_attrs__'):
                new_attrs = {}
                attrs = md['__xml_attrs__']
                for k, v in attrs.items():
                    if isinstance(k, unicode):
                        k = k.encode(enc)
                    if isinstance(v, unicode):
                        v = v.encode(enc)
                    new_attrs[k] = v
                md['__xml_attrs__'] = new_attrs
            self.values[idx] = name, ns, val, md
    return old_PropPatch_apply(self, obj)

old_Lock_apply = Lock.apply
def Lock_apply(self, obj, creator=None, depth='infinity',
               token=None, result=None, url=None, top=1):
    enc = site_encoding(obj, default='latin-1')
    # strval() has been patched to return unicode strings, so we
    # re-encode into site-encoding for the object.
    if self.owner and isinstance(self.owner, unicode):
        self.owner = self.owner.encode(enc)
    if token and isinstance(token, unicode):
        token = token.encode(enc)
    return old_Lock_apply(self, obj, creator=creator,
                          depth=depth, token=token,
                          result=result, url=url, top=top)

# Enable HTTP compression for dav requests.
old_resource_init = webdav.Resource.Resource.dav__init
def resource_dav__init(self, request, response):
    old_resource_init(self, request, response)
    response.enableHTTPCompression(request)

old_collection_init = webdav.Collection.Collection.dav__init
def collection_dav__init(self, request, response):
    old_collection_init(self, request, response)
    response.enableHTTPCompression(request)

def initialize(context):
    from App.version_txt import getZopeVersion, version_txt
    zope_version = getZopeVersion()
    str_version = version_txt().split()[1].strip(',')
    if zope_version > (2, 8, 1):
        log('Zope %s. Patches *NOT* applied.' % str_version)
    else:
        log('Zope %s. Applying patches.' % str_version)

        import webdav.xmltools
        import webdav.davcmds
        import xmltools
        webdav.xmltools.XmlParser = xmltools.XmlParser
        webdav.davcmds.XmlParser = xmltools.XmlParser
        log('Patch applied to webdav.xmltools.XmlParser.')

        from webdav.NullResource import NullResource
        NullResource.PUT = PUT
        log('Patch applied to webdav.NullResource.PUT.')

        from webdav.Resource import Resource
        Resource.COPY = COPY
        log('Patch applied to webdav.Resource.COPY.')

        from webdav.Resource import Resource
        Resource.MOVE = MOVE
        log('Patch applied to webdav.Resource.MOVE.')

        PropPatch.apply = PropPatch_apply
        log('Patch applied to PropPatch.apply.')

        Lock.apply = Lock_apply
        log('Patch applied to Lock.apply.')

        PropertySheets.xml_escape = xml_escape
        log('Patch applied to OFS.PropertySheets.xmlescape.')

        PropertySheets.PropertySheet.dav__allprop = dav__allprop
        log('Patch applied to OFS..PropertySheet.dav__allprop.')

        PropertySheets.PropertySheet.dav__propstat = dav__propstat
        log('Patch applied to OFS..PropertySheet.dav__propstat.')

        webdav.Resource.Resource.dav__init = resource_dav__init
        webdav.Collection.Collection.dav__init = collection_dav__init
        log('Patch applied to webdav.Collection/webdav.Resource dav__init.')

    try:
        from Products.Archetypes.utils import getPkgInfo
        import Products.Archetypes
        at_info = getPkgInfo(Products.Archetypes)
    except ImportError:
        log('Archetypes not found. *NOT* applying AT patches.')
    else:
        at_version = at_info.numversion
        str_version = at_info.version
        if at_version > (1, 3, 5, 0):
            log('Archetypes %s. Patches *NOT* applied.' % str_version)
        else:
            log('Archetypes %s. Applying patches.' % str_version)
            from Products.DavPack import at

    try:
        from Products.CMFPlone import PloneFolder
    except ImportError:
        log('Plone not found. *NOT* applying Plone patches.')
    else:
        # XXX needs to get proper Plone version somehow.
        plone_version = (0, 0, 0, 0)
        str_version = 'Unknown'
        if not plone_version < (2, 0, 5, 0):
            log('Plone %s. Patches *NOT* applied.' % str_version)
        else:
            log('Plone %s. Applying patches.' % str_version)
            from Products.DavPack import plone
